{Stack tracer

v0.14
25 March 2009

(C) Christian Ghisler, 2009
support@ghisler.com

History:
  v0.15 (7.7.09):
    + Fixed wrong start of trace (nearest try/except block instead of
      function with the bug)
    + Try to find Delphi stack frames
  v0.14 (25.3.09):
    + Removed all enable/disable functions
    + Changed code to work with Delphi 2
    + SYSTEM.PAS must be changed as follows:
      1. Under implementation add
       var RTLUnwindProc: Pointer;
      2. Replace all calls to RTLUnwind by calls to RTLUnwindProc
      3. In initialization code of unit, add
       RTLUnwindProc:=@RTLUnwind;
      4. In function BeginThread, add
       if StackSize=123456789 then begin
         result:=integer(RtlUnwindProc);
         RtlUnwindProc:=Parameter;
         exit;
       end;
       This is a hack so the interface section of system.pas remains unchanged

The contents of this file are subject to the Mozilla Public License
Version 1.1 (the "License"); you may not use this file except in
compliance with the License. You may obtain a copy of the License at
http://www.mozilla.org/MPL/

Software distributed under the License is distributed on an "AS IS"
basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See the
License for the specific language governing rights and limitations
under the License.

The Original Code is lk-stack-trace, available at
http://sourceforge.net/projects/lk-stack-trace/
http://orrollo.by.ru/
leon_kon@users.sourceforge.net

The Initial Developer of the Original Code is  Leonid Koninin
Portions created by the Initial Developer are Copyright (C) Leonid Koninin,
2007-2008. All Rights Reserved.

Contributor(s): Christian Ghisler

Alternatively, the contents of this file may be used under the terms
of the BSD license (the  "[BSD] License"), in which case the
provisions of [BSD] License are applicable instead of those
above. If you wish to allow use of your version of this file only
under the terms of the [BSD] License and not to allow others to use
your version of this file under the MPL, indicate your decision by
deleting the provisions above and replace them with the notice and
other provisions required by the [BSD] License. If you do not delete
the provisions above, a recipient may use your version of this file
under either the MPL or the [BSD] License.
}

{$OPTIMIZATION OFF}
{$STACKFRAMES ON}

unit STrace;

interface

uses
  windows,
  sysutils,
  classes;

var wantdebuginfo:boolean=false;        {Send stack trace debug info to OutputDebugString}

const
  MaxFilteredOutExceptions = 30;
  MAX_NFO_COUNT = 60; // max depth of stack tracing

function testcrash:boolean;
procedure ShowAppException(exceptiontext:pchar);

implementation

uses forms;

var exceptionbuf:array[0..1023] of char;
    lastexceptiontime:longint;

// ver 0.13+

procedure OurReporter(ws: pchar); stdcall;
var pstart,pend:pchar;
begin
  if not wantdebuginfo then exit;
  OutputDebugString(ws);
  exit;
  pstart:=ws;
  repeat
    pend:=strscan(pstart+1,#10);
    if pend<>nil then pend[0]:=#0;
    if pstart[0]<>#0 then
      OutputDebugString(pstart);
    if pend<>nil then pend[0]:=#10;
    pstart:=pend;
  until pstart=nil;
end;

// heuristic call check

function check_heuristic(xret: Cardinal): Boolean;

  function check1(xret: Cardinal): Boolean;
  begin
    result := not IsBadReadPtr(pointer(xret - 5), 5);
    if result then result := pbyte(xret - 5)^ = $E8;
  end;

  function check2(xret: Cardinal): Boolean;
  var
    pMasked: pWord;
    iBackLook: Integer;
  begin
    result := false;
    iBackLook := 2;
    while (not result) and (iBackLook < 8) do
      begin
        pMasked := pointer(xret - iBackLook);
        result := not IsBadReadPtr(pMasked, iBackLook);
        if result then result := (pMasked^ and $38FF) = $10FF;
        inc(iBackLook);
      end;
  end;

begin
  result := false;
  if not result then result := check1(xret);
  if not result then result := check2(xret);
end;

const HChar:array[0..15] of char='0123456789ABCDEF';

procedure LongToHex(Long:longint;Hex:PChar);
var i:integer;
begin
  i:=7;
  repeat
    Hex[i]:=HChar[Long and 15]; {Untere 4 Bits}
    Long:=Long shr 4; {div 16}
    dec(i);
  until i<0;
  Hex[8]:=#0
end;

procedure LongToHex6digit(Long:longint;Hex:PChar);
var i:integer;
begin
  i:=5;
  repeat
    Hex[i]:=HChar[Long and 15]; {Untere 4 Bits}
    Long:=Long shr 4; {div 16}
    dec(i);
  until i<0;
  Hex[6]:=#0
end;

// write context info to string and send it

type pCardinal=^Cardinal;
var max_stack, min_stack: Cardinal;

function validstackreturnaddress(prevbp:cardinal;recursive:boolean):boolean;
var cur_ebp,xret:Cardinal;
    i:integer;
    ok:boolean;
begin
  cur_ebp:=((prevbp-2) div 4) * 4 + 4;
  for i:=0 to 5 do begin   {Max. 6 parameters on the stack!}
    prevbp := pcardinal(cur_ebp-4)^;
    // check for stack frame!
    ok := (prevbp <= max_stack-16) and (prevbp >= min_stack) and (prevbp>cur_ebp);
    xret := pcardinal(cur_ebp)^;
    ok := ok and not IsBadCodePtr(pointer(xret));
    ok := ok and ((xret > $400000) and (xret < $800000));
    ok := ok and ((xret > max_stack) or (xret < min_stack));
    ok := ok and check_heuristic(xret);
    if recursive then
      ok := ok and validstackreturnaddress(prevbp,false);
    if ok then begin
      result:=true;
      exit
    end;
    inc(cur_ebp,4);
  end;
  result:=false;
end;

procedure trace_it(cur_ebp, ExceptionEIP, ExcCode, addinfo: Cardinal); stdcall;
var
  ok: Boolean;
  old_error: Cardinal;
  xret, ebp_of_except, prevbp: Cardinal;
  exc: TObject;
  nfo_count: Integer;
  firstcall:BOOL;
begin
  asm pushad; end;
  old_error := GetLastError;

  asm
    mov eax,fs:[18h];
    mov ecx,dword ptr [eax+4];
    mov max_stack,ecx;
    mov ecx,dword ptr [eax+8];
    mov min_stack,ecx;
  end;

  lastexceptiontime:=GetCurrentTime;  {Make sure we show the exception only when correlated}
  if (ExcCode = $0EEDFACE) and (addinfo <> 0) then begin
    exc := TObject(addinfo);
    strcopy(exceptionbuf,'Delphi exception: ');
    strpcopy(strend(exceptionbuf),exc.ClassName);
    strcat(exceptionbuf,#10);
  end else begin
    strcopy(exceptionbuf,'Windows exception: ');
    LongToHex(ExcCode,strend(exceptionbuf));
    strcat(exceptionbuf,#10);
  end;

  strcat(exceptionbuf,'Stack trace:'#10);
  LongToHex(ExceptionEIP,strend(exceptionbuf));
  strcat(exceptionbuf,#10);

  {Note: cur_ebp didn't point to location where error occurred, but to the function where
   the 'except' block was! -> Use current ebp!!!}
  ebp_of_except:=cur_ebp;

  asm
    mov cur_ebp, ebp
  end;
  inc(cur_ebp,1080);        {The first real stack value for a short function without params seems to be at 1088}
(*  strcat(exceptionbuf,'EBP:');
  LongToHex6digit(cur_ebp,strend(exceptionbuf));
  strcat(exceptionbuf,#10);*)

  nfo_count:=0;
  firstcall:=true;
  while (cur_ebp < max_stack) and (nfo_count < MAX_NFO_COUNT) do
    begin
      prevbp := pcardinal(cur_ebp-4)^;
      // check for stack frame!
      ok := (prevbp <= max_stack-16) and (prevbp >= min_stack) and (prevbp>cur_ebp);
      xret := pcardinal(cur_ebp)^;
      ok := ok and not IsBadCodePtr(pointer(xret));
      ok := ok and ((xret > $400000) and (xret < $800000));
      ok := ok and ((xret > max_stack) or (xret < min_stack));
      ok := ok and check_heuristic(xret);
      if firstcall then
        ok := ok and validstackreturnaddress(prevbp,true);
      if ok then
        begin
          if strlen(exceptionbuf)<sizeof(exceptionbuf)-8-2-1-10 then begin
            (*LongToHex6digit(cur_ebp,strend(exceptionbuf));
            strcat(exceptionbuf,':');*)

            if (ebp_of_except<>0) and (cur_ebp>=ebp_of_except) then begin
              strcat(exceptionbuf,'>');
              ebp_of_except:=0;
            end;
            LongToHex6digit(xret-2,strend(exceptionbuf));
            if nfo_count mod 6=5 then
              strcat(exceptionbuf,#10)
            else
              strcat(exceptionbuf,'  ');
          end;
          inc(nfo_count);
          cur_ebp:=((prevbp-2) div 4) * 4;
          firstcall:=false;
        end;
      inc(cur_ebp, 4);
    end;

  strlcat(exceptionbuf,#10'Raw:'#10,sizeof(exceptionbuf)-1);
  asm
    mov cur_ebp, ebp
  end;
  inc(cur_ebp,1080);
  nfo_count:=0;
  while (cur_ebp < max_stack) and (nfo_count < MAX_NFO_COUNT) do
    begin
      xret := pcardinal(cur_ebp)^;
      ok := not IsBadCodePtr(pointer(xret));
      ok := ok and ((xret > $400000) and (xret < $800000));
      ok := ok and ((xret > max_stack) or (xret < min_stack));
      ok := ok and check_heuristic(xret);
      if ok then
        begin
          if strlen(exceptionbuf)<sizeof(exceptionbuf)-8-2-1-10 then begin
            (*LongToHex6digit(cur_ebp,strend(exceptionbuf));
            strcat(exceptionbuf,':');*)

            if (ebp_of_except<>0) and (cur_ebp>=ebp_of_except) then begin
              strcat(exceptionbuf,'>');
              ebp_of_except:=0;
            end;
            LongToHex6digit(xret-2,strend(exceptionbuf));
            if nfo_count mod 6=5 then
              strcat(exceptionbuf,#10)
            else
              strcat(exceptionbuf,'  ');
          end;
          inc(nfo_count);
        end;
      inc(cur_ebp, 4);
    end;


  OurReporter(exceptionbuf);

  SetLastError(old_error);
  asm popad; end;
end;

var
  oldUnWindProc: procedure stdcall = nil;

function IsDelphiCode(nn: Cardinal): Boolean; stdcall;
begin
  result := nn = $0EEDFADE;
  result := result or (nn = $0EEDFAE0);
end;

// prototype of RtlUnwind is from MSDN

procedure LkRtlUnwind(TargetFrame, TargetIp, ExceptionRecord,
  ReturnValue: Pointer); stdcall;
asm
  pushad;
  mov       eax,[ebp+8];
  or        eax,eax;
  jz        @log_skip;
  mov       ebx,[eax+8];
  mov       eax,[ebp+16];
  push      dword ptr [eax+24];
  push      dword ptr [eax];
  push      dword ptr [eax];
  call      IsDelphiCode;
  or        al,al;
  mov       eax,[ebp+16];
  jnz       @delphi_exception;
  mov       eax,[eax+12];
  push      eax;
  jmp       @make_trace;
@delphi_exception:
  mov       eax,[eax+20];
  push      eax;
@make_trace:
  push      ebx;
@go_trace:
  call      trace_it;
  jmp       @oldproc;
// on no frame
@log_skip:
  nop;
// jump to old code
@oldproc:
  popad;
  pop       ebp;
  jmp       oldUnWindProc;
end;

procedure hook_unwind;
var DummyThreadId:integer;
begin
  {See needed changes for SYSTEM.PAS above!!!}
  OldUnwindProc:=tfarproc(BeginThread(nil,123456789,nil,@LkRtlUnwind,0,DummyThreadId));
end;

var p:pwidechar;
function test2:boolean;
begin
  p:=nil;
  result:=p[0]=#0;
end;

function test1(bla:integer):boolean; stdcall;
var a:integer;
begin
  a:=bla;
  result:=test2 and (a=0);
end;

function testcrash:boolean;
begin
  p:=nil;
  result:=p[0]=#0;
//  result:=test1(0);
end;

var crashdepth:integer=0;

procedure ShowAppException(exceptiontext:pchar);
const s1:pchar=#10#10'Please report this error to the Author, with a description'#10'of what'+
               ' you were doing when this error occurred!';
var msgbuf:array[0..1511] of char;
    ch:char;
begin
  if crashdepth>100 then halt(1);  {No error when quitting}
  if crashdepth>50 then messagebeep(0);
  inc(crashdepth);
  if crashdepth>1 then exit;     {Avoid multiple nested dialog boxes!}
  strlcopy(msgbuf,exceptiontext,sizeof(msgbuf)-1);
  if (msgbuf[0]<>#0) then begin
    ch:=msgbuf[strlen(msgbuf)-1];
    if (ch<>'!') and (ch<>'.') then
      strlcat(msgbuf,'.',sizeof(msgbuf)-1);
  end;
  strlcat(msgbuf,s1,sizeof(msgbuf)-1);
  if (exceptionbuf[0]<>#0) and (abs(GetCurrentTime-lastexceptiontime)<5000) then begin
    exceptionbuf[sizeof(exceptionbuf)-1]:=#0;
    strlcat(msgbuf,#10#10,sizeof(msgbuf)-1);
    strlcat(msgbuf,exceptionbuf,sizeof(msgbuf)-1);
    exceptionbuf[0]:=#0;
  end;
  strlcat(msgbuf,#10'Press Ctrl+C to copy this report!'#10'Continue execution?',sizeof(msgbuf)-1);

  try
    if strpos(msgbuf,'SplitPanel.Cursor')=nil then {Ignore SplitPanel.Cursor load error}
      if MessageBox(getactivewindow,msgbuf,'Error!', MB_YESNO + MB_ICONSTOP)=idNo then
        application.terminate;
  except
    crashdepth:=101;
    halt(1);
  end;
end;

initialization
  exceptionbuf[0]:=#0;
  hook_unwind;

finalization
end.

